class EventTypes:
    SERVED_CONNECTIONS = 'NAMES'
    USER_JOINED = 'JOIN'
    USER_QUIT = 'QUIT'
    BROADCAST = 'BROADCAST'
    CHAT_MESSAGE = 'MSG'
    WHISPER = 'PRIVMSG'
    WHISPER_SENT = 'PRIVMSGSENT'
    MUTE = 'MUTE'
    UNMUTE = 'UNMUTE'
    BAN = 'BAN'
    UNBAN = 'UNBAN'
    SUB_ONLY = 'SUBONLY'
    ERROR = 'ERR'

    class Special:
        ANY_MESSAGE = 'ANY_MESSAGE'
        MENTION = 'MENTION'
        WS_ERROR = 'WS_ERROR'
        WS_CLOSE = 'WS_CLOSE'

    @staticmethod
    def is_moderation_event(msg_type):
        return msg_type in (EventTypes.MUTE, EventTypes.UNMUTE, EventTypes.BAN, EventTypes.UNBAN)



import logging
from re import findall
from json import loads
from requests import get

from ..exceptions import APIError, CDNError


class DGGCDN:
    """
    An API for dgg's CDN.
    Let's you get info about emotes, flairs, and maybe other stuff.
    """

    DGG = 'https://www.destiny.gg/embed/chat'
    PATTERN = r'data-cache-key=\"(\d+\.\d+)\" data-cdn=\"(https?:\/\/(?:www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b(?:[-a-zA-Z0-9()@:%_\+.~#?&//=]*))\"'

    def __init__(self):
        self.cdn = None
        self.cache_key = None

        self._setup()

    def _setup(self):
        logging.info('setting up cdn...')

        r = get(self.DGG)

        if r.status_code != 200:
            raise APIError(DGG, r)

        content = r.content.decode('utf8')
        match = findall(self.PATTERN, content)

        if not match:
            raise CDNError('failed to find cache key and cdn url')

        self.cache_key, self.cdn = match[0]

        logging.info(f"setup cdn: {self.cdn} cache key: {self.cache_key}")

    def get(self, object_path, as_json=True):
        if not object_path.startswith('/'):
            object_path = f"/{object_path}"

        url = f"{self.cdn}{object_path}?_={self.cache_key}"
        logging.info(f"retrieving: {url}")
        r = get(url, allow_redirects=False)

        logging.info(f"retrieved: {r.content.decode('utf8')[:200]}")

        if r.status_code in (301, 404):
            raise FileNotFoundError(object_path)

        if r.status_code != 200:
            raise APIError(url, r)

        if as_json:
            return loads(r.content.decode('utf8'))

        return r.content

    def get_flairs(self):
        return self.get('/flairs/flairs.json')

    def get_emotes(self):
        return self.get('/emotes/emotes.json')



from ._exceptions import (
    APIError,
    AnonymousConnectionError,
    AnonymousSessionError,
    DumbFucksBeware,
    CDNError,
    InvalidAuthTokenError,
    InvalidChatLine,
    InvalidMessageError,
)



class StreamInfo:
    def __init__(
        self,
        started_at, ended_at, duration,
        viewers, game, host, live,
        preview, status_text
    ):
        self.started_at = started_at
        self.ended_at = ended_at
        self.duration = duration
        self.viewers = viewers
        self.game = game
        self.host = host
        self.live = live
        self.preview = preview
        self.status_text = status_text

    @classmethod
    def from_api_response(cls, response):
        return cls(**response)



from json import dumps
from datetime import datetime

from ..messages import Message
from .._utils import dict_swap_keys
from .._event_types import EventTypes


class PrivateMessage:
    def __init__(
        self, id,
        deleted_by_receiver,
        deleted_by_sender,
        from_user, target_user,
        from_user_id, target_user_id,
        date_time, is_read, content
    ):
        self.id = id
        self.deleted_by_receiver = bool(int(deleted_by_receiver))
        self.deleted_by_sender = bool(int(deleted_by_sender))
        self.from_user = from_user
        self.target_user = target_user
        self.from_user_id = from_user_id
        self.target_user_id = target_user_id
        self.date_time = datetime.strptime(date_time, '%Y-%m-%dT%H:%M:%S+0000')
        self.is_read = bool(int(is_read))
        self.content = content

    @classmethod
    def from_api_response(cls, response):
        key_map = {
            'deletedbyreceiver': 'deleted_by_receiver',
            'deletedbysender': 'deleted_by_sender',
            'from': 'from_user',
            'to': 'target_user',
            'userid': 'from_user_id',
            'targetuserid': 'target_user_id',
            'isread': 'is_read',
            'message': 'content',
            'timestamp': 'date_time',
        }
        response = dict_swap_keys(response, key_map)
        return cls(**response)

    @property
    def as_websocket_message(self):
        payload = {
            'messageid': self.id,
            'nick': self.from_user,
            'timestamp': int(1000*self.date_time.timestamp()),
            'data': self.content,
        }
        msg = f"{EventTypes.WHISPER} {dumps(payload, ensure_ascii=False)}"
        return Message.parse(msg)



import logging
from typing import Union
from requests import get
from datetime import datetime

from ._chat_message import ChatMessage

from ..exceptions import APIError
from .._utils import validate_date_time, isplit


class DGGLogs:
    """
    API for OverRustleLogs dgg logs.
    API calls return generators, thus memory efficient.
    Messages are timestamped in UTC.
    """

    DGG_LOGS = 'https://overrustlelogs.net/Destinygg%20chatlog'

    @classmethod
    def _parse_logs(cls, logs):
        for line in isplit(logs, '\n'):
            if line:
                yield ChatMessage.from_chat_line(line)

    @classmethod
    def _build_date(cls, year=0, month=None):
        if not month:
            month = datetime.utcnow().month
        if not year:
            year = datetime.utcnow().year

        if isinstance(month_full := month, int):
            month_full = datetime.strptime(str(month), '%m').strftime('%B')

        # raises `ValueError`
        validate_date_time(year=year, month=month)
        return f"{month_full} {year}"

    @classmethod
    def _get(cls, endpoint, year=0, month=None):
        date = cls._build_date(year, month)
        url = f"{cls.DGG_LOGS}/{date}/{endpoint}"

        logging.info(f"retrieving: {url}")

        r = get(url)

        logging.info(f"response: {r.status_code}")

        if r.status_code == 404:
            raise FileNotFoundError

        if r.status_code != 200:
            raise APIError(url, r)

        return cls._parse_logs(r.content.decode('utf8'))

    @classmethod
    def get_daily_logs(cls, year=0, month=0, day=0):
        """
        Retrieves chat logs for a specific day, month, and year.
        If any isn't provided, use current based on UTC time.
        """

        now = datetime.utcnow()

        year = year or now.year
        month = month or now.month
        day = day or now.day

        date = f"{year}-{month:02d}-{day:02d}"

        endpoint = f"{date}.txt"
        return cls._get(endpoint, year, month)

    @classmethod
    def get_user_logs(cls, user, year=0, month: Union[int, str] = None):
        """
        Retrieves user logs for a specific month and year.
        If either is not provided, use current based on UTC time.

        `month` : `int` or `str`

            if `str`, has to be the full month name ('January', 'February', ...).

        """

        endpoint = f"userlogs/{user}.txt"
        return cls._get(year, month, endpoint)

    @classmethod
    def get_broadcaster_logs(cls, year=0, month: Union[int, str] = None):
        """
        Retrieves broadcaster logs for a specific month and year.
        If either is not provided, use current based on UTC time.

        `month` : `int` or `str`

            if `str`, has to be the full month name ('January', 'February', ...).

        """

        endpoint = f"broadcaster.txt"
        return cls._get(endpoint, year, month)

    @classmethod
    def get_subscribers(cls, year=0, month: Union[int, str] = None):
        """
        Retrieves users who subscribed in a specific month and year.
        If either is not provided, use current based on UTC time.

        `month` : `int` or `str`

            if `str`, has to be the full month name ('January', 'February', ...).

        """

        endpoint = f"subscribers.txt"
        return cls._get(endpoint, year, month)

    @classmethod
    def get_bans(cls, year=0, month: Union[int, str] = None):
        """
        Retrieves ban messages from a specific month and year.
        If either is not provided, use current based on UTC time.

        `month` : `int` or `str`

            if `str`, has to be the full month name ('January', 'February', ...).

        """

        endpoint = f"bans.txt"
        return cls._get(endpoint, year, month)



import os
import logging
from logging import NOTSET, DEBUG, INFO, WARNING, ERROR, CRITICAL
from logging.handlers import RotatingFileHandler


MAX_FILE_SIZE = 10*int(2**20)  # 10MiB


def setup_logger(level=logging.NOTSET, file_name='dgg.log', log_to_console=True):
    if not os.path.exists('logs'):
        os.mkdir('logs')
    handler = RotatingFileHandler(
        f"logs/{file_name}", maxBytes=MAX_FILE_SIZE,
        backupCount=10, encoding='utf8'
    )
    handlers = [handler]
    if log_to_console:
        handlers.append(logging.StreamHandler())
    logging.basicConfig(
        handlers=handlers,
        level=level,
        format=u"[%(asctime)s.%(msecs)03d] %(levelname)s [%(name)s.%(funcName)s:%(lineno)d] %(message)s",
        datefmt='%Y-%m-%dT%H:%M:%S'
    )
    logging._defaultFormatter = logging.Formatter(u"%(message)s")



from re import compile
from json import dumps
from types import MethodType
from datetime import datetime


_CAMEL_SNAKE_PATTERN = compile(r'(?<!^)(?=[A-Z])')


def format_payload(type, **kwargs):
    """Formats a message to the DGG websocket format."""
    return f"{type} {dumps(kwargs)}"


def dict_swap_keys(d, key_map):
    swapped = d.copy()
    for key in d:
        if key in key_map:
            new_key = key_map[key]
            swapped[new_key] = swapped.pop(key)
    return swapped


def camel_to_snake_case(s):
    return _CAMEL_SNAKE_PATTERN.sub('_', s).lower()


def dict_keys_camel_to_snake_case(d):
    snaked = d.copy()
    for key in d:
        new_key = camel_to_snake_case(key)
        snaked[new_key] = snaked.pop(key)
    return snaked


def format_datetime(dt, fmt='%Y-%m-%d %H:%M:%S'):
    return dt.strftime(fmt)


def validate_date_time(year=1, month=1, day=1, hour=0, minute=0, second=0, microsecond=0):
    datetime(year, month, day, hour, minute, second, microsecond)


def isplit(source, sep=None, regex=False):
    """
    https://stackoverflow.com/a/9773142
    generator version of str.split()

    :param source:
        source string (unicode or bytes)

    :param sep:
        separator to split on.

    :param regex:
        if True, will treat sep as regular expression.

    :returns:
        generator yielding elements of string.
    """
    if sep is None:
        # mimic default python behavior
        source = source.strip()
        sep = "\\s+"
        if isinstance(source, bytes):
            sep = sep.encode("ascii")
        regex = True
    if regex:
        # version using re.finditer()
        if not hasattr(sep, "finditer"):
            sep = re.compile(sep)
        start = 0
        for m in sep.finditer(source):
            idx = m.start()
            assert idx >= start
            yield source[start:idx]
            start = m.end()
        yield source[start:]
    else:
        # version using str.find(), less overhead than re.finditer()
        sepsize = len(sep)
        start = 0
        while True:
            idx = source.find(sep, start)
            if idx == -1:
                yield source[start:]
                return
            yield source[start:idx]
            start = idx + sepsize



from datetime import datetime

from .._utils import dict_keys_camel_to_snake_case


class User:
    def __init__(
        self, id=None, nick=None, username=None,
        status=None, auth_provider=None, country=None,
        created_date=None, features=None, roles=None,
        settings=None, subscription=None
    ):
        self.id = id
        self.nick = nick
        self.username = username
        self.status = status
        self.auth_provider = auth_provider
        self.country = country  # currently not supported
        self.created_date = datetime.strptime(
            created_date, '%Y-%m-%dT%H:%M:%S+0000'
        ) if created_date else None
        self.features = features
        self.roles = roles
        self.settings = settings
        self.subscription = subscription

    def __repr__(self):
        return f"User(id='{self.id}', nick='{self.nick}')"

    @classmethod
    def from_api_response(cls, response):
        response = dict_keys_camel_to_snake_case(response)
        if 'user_id' in response:
            response['id'] = response.pop('user_id')
        if 'user_status' in response:
            response['status'] = response.pop('user_status')
        return cls(**response)

    @property
    def is_subbed(self):
        return "subscriber" in self.features



from ._messages import (
    EventTypes,
    ChatUser,
    Message,
    ServedConnections,
    UserJoined,
    UserQuit,
    Broadcast,
    ChatMessage,
    Whisper,
    ModerationMessage,
    SubOnly,
)



from re import findall
from json import dumps
from datetime import datetime

from ..exceptions import InvalidChatLine
from .._utils import format_datetime


class ChatMessage:
    CHAT_LINE_PATTERN = r'^\[(\d{4}-\d\d-\d\d \d\d:\d\d:\d\d) UTC\] (\w+): (.+)$'

    def __init__(self, user, date_time: datetime, content):
        self.user = user
        self.date_time = date_time
        self.content = content

    def __repr__(self):
        return f"ChatMessage(user='{self.user}', date_time='{format_datetime(self.date_time)}', content='{self.content}')"

    @classmethod
    def from_chat_line(cls, line):
        match = findall(cls.CHAT_LINE_PATTERN, line.strip())
        if not match:
            raise InvalidChatLine(line)

        date_time, user, content = match[0]
        return cls(
            user,
            datetime.strptime(date_time, '%Y-%m-%d %H:%M:%S'),
            content,
        )

    @property
    def is_from_anonymous_user(self):
        return self.user == '_anon$'



import time
import logging
import queue
from os import getenv
from queue import Queue
from threading import Thread
from websocket import WebSocketApp
from websocket._exceptions import WebSocketException

from .exceptions import (
    AnonymousConnectionError,
    AnonymousSessionError,
    DumbFucksBeware,
    InvalidAuthTokenError,
    InvalidMessageError,
)
from .messages import Message, EventTypes, ChatUser
from .api import DGGAPI, User
from ._utils import format_payload
from ._handler import DGGChatEventHandler


class DGGChat:
    """
    A dgg chat API.    
    """

    DGG_WS = 'wss://destiny.gg/ws'
    RECONNECT_DELAY = 3    # in seconds
    WAIT_WS_BOOTSTRAP = 1  # in seconds
    # considering server throttle is 300ms and some network overhead
    WS_THROTTLE_DELAY = .200  # in seconds
    # after this much time after last sent message, throttle factor is reset
    WS_THROTTLE_RESET = 600  # in seconds
    # throttling should be a very exceptional case, so instead of starting at 1,
    # a bit of padding ensures it doesn't happen multiple times in a row
    BASE_WS_THROTTLE_FACTOR = 1.1

    def __init__(
        self, auth_token=None, session_id=None,
        validate_auth_token=True,
        handle_history=False,
        handle_unread_whispers=False,
        mark_as_read=False,
        anti_throttle_bot='',
    ):
        """
        Parameters
        ----------
        `auth_token` : `str`

            an authentication token for a dgg account.
            can be created at `https://www.destiny.gg/profile/developer`.

        `session_id` : `str`

            a session id for a dgg connection. used for profile related stuff.
            currently needs to be set manually by logging in on the browser and copying the cookie.

        `validate_auth_token` : `bool`

            whether the token should be validated. `True` by default.
            (should only be needed to disable it if something's changed in how tokens are generated)

        `handle_unread_whispers` : `bool`

            whether unread whispers should be handled. requires `session_id` to work.
            handlers are called upon connecting (if `handle_history` also enabled, this happens before).
            `False` by default.

        `handle_history` : `bool`

            whether the previous 150 most recent messages from before connection should be handled.
            handlers are called upon connecting (if `handle_unread_whispers` also enabled, this happens after).
            `False` by default.

        `mark_as_read` : `bool`

            whether after handling a whisper it should be marked as read in the chat backend.
            doing so will make it so it doesn't show up when calling `get_unread_whispers()`.
            requires `session_id` to work.
            `False` by default.

        `anti_throttle_bot` : `str`

            the chat nick for a echo bot connected in chat.
            used to bypass throttling limits.
        """

        if auth_token and validate_auth_token and not self.auth_token_is_valid(auth_token):
            raise InvalidAuthTokenError(auth_token)

        self.mark_as_read = mark_as_read
        self.handle_history = handle_history
        self.handle_unread_whispers = handle_unread_whispers

        self._auth_token = auth_token
        self._session_id = session_id

        self._running = False
        self._users_available_to_whisper = set()

        self._last_message_time = 0
        self._next_message_time = time.time()
        self._ws_throttle_factor = self.BASE_WS_THROTTLE_FACTOR
        self._anti_throttle_bot = anti_throttle_bot

        self._queued_messages = Queue()
        self._unhandled_messages = Queue()

        self._handler = DGGChatEventHandler()
        self._api = DGGAPI(auth_token, session_id)

        self._profile = self._reload_profile() if auth_token else None

        self._setup_web_socket()

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, type, value, traceback):
        self.disconnect()

    @property
    def throttle_factor(self):
        return self._ws_throttle_factor

    @property
    def profile(self):
        return self._profile

    @property
    def api(self):
        return self._api

    @staticmethod
    def auth_token_is_valid(token):
        return len(token) == 64 and token.isalnum()

    @staticmethod
    def message_is_valid(msg):
        return 0 < len(msg) <= 512

    def _setup_web_socket(self):
        def on_message(ws, message):
            """The top-level message handling function."""

            parsed = Message.parse(message)

            logging.debug(f"received message: `{message}`")
            logging.info(f"parsed message: `{parsed}`")

            if not self._unhandled_messages.empty() and parsed.event in (EventTypes.ERROR, EventTypes.WHISPER_SENT):
                self._handle_message(parsed)

            if parsed.event == EventTypes.WHISPER:
                enabled = str(getenv('DGG_ENABLE_WHISPERS')).lower()
                if enabled and enabled != 'false':
                    self._users_available_to_whisper.add(parsed.user.nick)
                    logging.info(
                        f"{parsed.user.nick} added to users available to whisper"
                    )

            self._handler.handle_event(EventTypes.Special.ANY_MESSAGE, parsed)

            if self._profile and parsed.event == EventTypes.CHAT_MESSAGE and self._profile.nick in parsed.content:
                self._handler.handle_event(EventTypes.Special.MENTION, parsed)

            if parsed.event == EventTypes.WHISPER and self.mark_as_read:
                self.mark_all_as_read(parsed.user.nick)

            if parsed.event == EventTypes.WHISPER_SENT:
                # `on_whisper_sent` handler takes no arguments
                return self._handler.handle_event(parsed.event)
            self._handler.handle_event(parsed.event, parsed)

        def on_ws_error(ws, error):
            """Handler for websocket related errors."""

            self._handler.handle_event(EventTypes.Special.WS_ERROR, error)

            msg = f"websocket error: `{error}`"

            logging.error(msg)
            raise WebSocketException(msg)

        def on_close(ws):
            """Handler for when the websocket connection is closed."""

            self._handler.handle_event(EventTypes.Special.WS_CLOSE)
            self._running = False

            logging.info('connection closed')

        self._ws = WebSocketApp(
            self.DGG_WS,
            on_message=on_message,
            on_error=on_ws_error,
            on_close=on_close,
            cookie=f"authtoken={self._auth_token}" if self._auth_token else None
        )

    def _reload_profile(self):
        self._profile = self._api.user_info()
        logging.info(f"profile updated: {self._profile}")

    def _handle_history(self):
        """Handles the 150 most recent chat messages up until chat is connected."""

        logging.info('handling history')

        history = self._api.chat_history()
        for message in history:
            self._handler.handle_event(message.event, Message.parse(message))

    def _handle_unread_whispers(self):
        """Handles unread whispers. Marks them as read."""

        if not self._session_id:
            raise AnonymousConnectionError('unable to handle unread whispers')

        logging.info('handling unread whispers')

        unread = self.get_unread_whispers()
        for user, whispers in unread.items():
            for whisper in whispers:
                self._handler.handle_event(
                    whisper.event, whisper.as_websocket_message
                )

    def _handle_message(self, message):
        try:
            payload = self._unhandled_messages.get_nowait()
        except queue.Empty as e:
            logging.fatal('unhandled messages queue was empty')
            raise e

        now = time.time()
        if now >= self._last_message_time + self.WS_THROTTLE_RESET:
            logging.info('resetting throttle factor')
            self._ws_throttle_factor = self.BASE_WS_THROTTLE_FACTOR

        if message.event == EventTypes.ERROR:
            # max throttle factor seems to be 16 (16*.3=5s)
            # verified empirically since this doesn't seem to match the source code (https://github.com/destinygg/chat/blob/master/connection.go#L407)
            if message.payload == 'throttled':
                logging.warning('connection throttled')
                self._ws_throttle_factor = min(16, 2*self._ws_throttle_factor)

            if message.payload == 'duplicate':
                logging.warning('duplicate message')
                self._ws_throttle_factor = min(
                    16, 1 + self._ws_throttle_factor
                )
        else:
            self._last_message_time = now

        self._next_message_time = now + self._ws_throttle_factor*self.WS_THROTTLE_DELAY

    def _start_send_loop(self):
        Thread(target=self._send_loop, daemon=True).start()

    def _send_loop(self):
        time.sleep(self.WAIT_WS_BOOTSTRAP)
        self._running = True
        while True:
            while not self._next_message_time or time.time() < self._next_message_time:
                continue

            if not self._running:
                break

            self._next_message_time = 0

            payload = self._queued_messages.get()
            logging.debug(f"sending payload: `{payload}`")

            self._ws.send(payload)
            self._unhandled_messages.put(payload)

            if self._anti_throttle_bot:
                anti_throttle_payload = format_payload(
                    EventTypes.WHISPER, nick=self._anti_throttle_bot, data='0'
                )
                logging.debug(
                    f"anti-throttle payload: `{anti_throttle_payload}`"
                )

                self._ws.send(anti_throttle_payload)
                self._unhandled_messages.put(anti_throttle_payload)

    def _queue_message(self, type, **kwargs):
        payload = format_payload(type, **kwargs)
        logging.debug(f"enqueueing payload: `{payload}`")
        self._queued_messages.put(payload)

    def update_profile(self):
        """
        Updates `self.profile` with info from the auth token's account.
        Useful on flair or subscriber status changes.
        """

        if not self._auth_token and not self._session_id:
            raise AnonymousConnectionError('unable to update profile')

        self._reload_profile()

    def mark_all_as_read(self, from_user=None):
        # TODO: implement more efficiently (use `/messages/inbox|read` endpoints)
        self.get_unread_whispers(from_user=from_user)

    def get_unread_whispers(self, from_user=None, received_only=True):
        """
        Retrieves all unread whispers. Includes the messages sent to the user.
        When `received_only` is `True`, messages sent are filtered out,
        returning only those received. All messages are marked as read.
        """

        if not self._session_id:
            raise AnonymousSessionError('unable to get unread messages')

        unread = self._api.messages_unread()
        messages = {}

        if from_user:
            if from_user in unread:
                messages[from_user] = self._api.messages_inbox(
                    from_user, unread[from_user]
                )
            else:
                messages[from_user] = []
            return messages

        for user, count in unread.items():
            messages[user] = self._api.messages_inbox(user, count)

        return messages

    def connect(self):
        """Connect to chat and run in a new thread (non-blocking)."""

        if self._running:
            raise ConnectionError('chat is already connected')

        logging.info('setting up connection')
        t = Thread(target=self.run_forever, daemon=True)
        t.start()

        time.sleep(self.WAIT_WS_BOOTSTRAP)
        logging.info('connected')
        return t

    def disconnect(self):
        if not self._running:
            raise ConnectionError('chat is not connected')

        logging.info('disconnecting')
        self._ws.close()

        logging.info('disconnected')
        self._running = False

    def run_forever(self):
        """Connect to chat and block the thread."""

        if self._running:
            raise ConnectionError(
                'chat already connected, call `disconnect()` first'
            )
        
        if self.handle_unread_whispers:
            self._handle_unread_whispers()

        if self.handle_history:
            self._handle_history()

        logging.info('running websocket on loop')
        self._start_send_loop()

        while self._ws.run_forever():
            logging.warning(
                f"connection dropped. trying to reconnect in {self.RECONNECT_DELAY} seconds..."
            )
            time.sleep(self.RECONNECT_DELAY)

    def send_whisper(self, user, message):
        if not self.message_is_valid(message):
            raise InvalidMessageError(message)

        if not self._auth_token and not self._session_id:
            raise AnonymousConnectionError('unable to send whispers')

        if self._profile and self._profile.nick == user:
            raise ValueError("you can't whisper yourself, you silly goose")

        enabled = str(getenv('DGG_ENABLE_WHISPERS')).lower()
        if not enabled or enabled in ('none', 'false') or user not in self._users_available_to_whisper:
            raise DumbFucksBeware('cannot send whispers')

        logging.info('queue send whisper')
        self._queue_message(EventTypes.WHISPER, nick=user, data=message)

    def _on(self, f, event):
        return self._handler.on(f, event)

    def on_any_message(self, f):
        """
        Called when receiving any message. 
        Specific handler still called as usual.
        """

        return self._on(f, EventTypes.Special.ANY_MESSAGE)

    def on_served_connections(self, f):
        """
        Called when receiving the first message when a new connection is established,
        which lists all users connected and amount of connections currently served.
        """

        return self._on(f, EventTypes.SERVED_CONNECTIONS)

    def on_user_joined(self, f):
        return self._on(f, EventTypes.USER_JOINED)

    def on_user_quit(self, f):
        return self._on(f, EventTypes.USER_QUIT)

    def on_broadcast(self, f):
        """
        Called when receiving broadcasts (the yellow messages),
        such as when a user subscribes.
        """

        return self._on(f, EventTypes.BROADCAST)

    def on_chat_message(self, f):
        return self._on(f, EventTypes.CHAT_MESSAGE)

    def on_mention(self, f):
        """
        Called when a chat message contains the current user's name.
        It's not called by the handler, but by the `DGGChat` instance,
        so it doesn't need to be mapped. `on_chat_message()` is still called.
        """

        return self._on(f, EventTypes.Special.MENTION)

    def on_whisper(self, f):
        return self._on(f, EventTypes.WHISPER)

    def on_whisper_sent(self, f):
        """Called on confirmation messages that a whisper was successfully sent."""

        return self._on(f, EventTypes.WHISPER_SENT)

    def on_mute(self, f):
        return self._on(f, EventTypes.MUTE)

    def on_unmute(self, f):
        return self._on(f, EventTypes.UNMUTE)

    def on_ban(self, f):
        return self._on(f, EventTypes.BAN)

    def on_unban(self, f):
        return self._on(f, EventTypes.UNBAN)

    def on_sub_only(self, f):
        return self._on(f, EventTypes.SUB_ONLY)

    def on_error(self, f):
        """
        Called on an error message when something goes wrong, 
        such as when sending a whisper to a user that doesn't exist.
        """

        return self._on(f, EventTypes.ERROR)

    def on_ws_error(self, f):
        """
        Called when something goes wrong with the websocket connection.
        It's not called by the handler, but by the `DGGChat` instance,
        so it doesn't need to be mapped.
        """

        return self._on(f, EventTypes.Special.WS_ERROR)

    def on_ws_close(self, f):
        """
        Called when the websocket connection is closed.
        It's not called by the handler, but by the `DGGChat` instance,
        so it doesn't need to be mapped.
        """

        return self._on(f, EventTypes.Special.WS_CLOSE)



import logging


class APIError(Exception):
    def __init__(self, endpoint, response):
        message = f"failed to call `{endpoint}`: {response.status_code} `{response.content.decode('utf8')}`"
        logging.warning(message)
        super().__init__(message)


class AnonymousConnectionError(Exception):
    def __init__(self, message='connection is anonymous'):
        message = f"{message}: no auth token provided"

        logging.warning(message)

        super().__init__(message)


class AnonymousSessionError(Exception):
    def __init__(self, message='session is anonymous'):
        message = f"{message}: no session id provided"

        logging.warning(message)

        super().__init__(message)


class CDNError(Exception):
    def __init__(self, message='CDN error'):
        super().__init__(message)


class DumbFucksBeware(Exception):
    def __init__(self, message):
        super().__init__(
            f"{message}\n"
            "If you end up doing some dumb shit and getting banned from dgg, "
            "I take no responsibility from this point on. "
            "In any case, have a look at the source code to get what you need."
        )


class InvalidAuthTokenError(Exception):
    def __init__(self, auth_token, message=''):
        message = message or (
            f"invalid auth token `{auth_token}`. it should be a 64 character alphanumeric string. "
            'if you believe this is a mistake, set `validate_auth_token` to `False`'
        )
        super().__init__(message)


class InvalidChatLine(Exception):
    def __init__(self, line):
        super().__init__(line)


class InvalidMessageError(Exception):
    def __init__(self, message=''):
        message = message or 'message length should be in inclusive range [1, 512]'

        logging.warning(message)

        super().__init__(message)



import logging
from json import loads
from datetime import datetime
from requests import get

from ._user import User
from ._stream_info import StreamInfo
from ._private_message import PrivateMessage
from ..exceptions import APIError, AnonymousConnectionError, AnonymousSessionError


class DGGAPI:
    DGG_API = 'https://www.destiny.gg/api'

    def __init__(self, auth_token=None, session_id=None):
        self._auth_token = auth_token
        self._session_id = session_id
        self._profile: User = None

    @property
    def auth_token(self):
        return self._auth_token

    def _get(self, endpoint, as_json=True):
        logging.info(f"calling api on `{endpoint}`")

        endpoint = f"{self.DGG_API}{endpoint}"
        cookies = dict(authtoken=self._auth_token, sid=self._session_id)

        r = get(endpoint, cookies=cookies)
        if r.status_code != 200:
            raise APIError(endpoint, r)

        logging.info(f"received from api: `{r.content.decode('utf8')}`")
        return loads(r.content) if as_json else r.content

    def user_info(self):
        if not self._auth_token:
            raise AnonymousConnectionError('unable to get profile')
        user = self._get(f"/userinfo?token={self._auth_token}")
        self._profile = User.from_api_response(user)
        return self._profile

    def chat_me(self):
        if not self._auth_token:
            raise AnonymousSessionError('unable to get profile')

        user = self._get('/chat/me')
        self._profile = User.from_api_response(user)
        return self._profile

    def chat_history(self):
        """
        Returns a list of the last 150 chat messages from right before connecting.
        The format is the same as used in the WebSocket interface.
        """

        return self._get('/chat/history')

    def messages_unread(self):
        """
        Returns a dictionary with the username and the amount of private messages
        currently unread from that user.
        """

        if not self._session_id:
            raise AnonymousSessionError('unable to get unread messages')

        unread = self._get('/messages/unread')
        return {
            m.get('username'): int(m.get('unread'))
            for m in unread
        }

    def _get_inbox(self, user, offset=0, received_only=True):
        inbox = self._get(f"/messages/usr/{user}/inbox?s={offset}")

        messages = []
        for m in inbox:
            pm = PrivateMessage.from_api_response(m)
            # should be false when `pm_is_read` OR when `received_only` AND `nick` != `target`
            if not pm.is_read and not (received_only and self._profile.nick != pm.target_user):
                messages.append(pm)
        return messages

    def messages_inbox(self, user, count=25, offset=0, received_only=True):
        """
        Returns a list of the last `count` `PrivateMessages` exchanged with `user`.
        `offset` can be user to skip to older messages.
        This marks messages as read.
        """

        if not self._session_id:
            raise AnonymousSessionError('unable to get inbox')

        messages = []
        while True:
            inbox = self._get_inbox(user, offset, received_only=True)
            if not len(inbox):
                break
            messages.extend(inbox[:count])
            if len(inbox) >= count:
                break
            count -= len(inbox)
            offset += len(inbox)

        return messages

    def info_stream(self):
        """Returns info about current stream if live, or last stream."""

        info = self._get('/info/stream')
        return StreamInfo.from_api_response(info)



import logging


class DGGChatEventHandler:
    def __init__(self):
        self._handlers = {}

    def on(self, f, event):
        self._handlers.setdefault(event, set()).add(f)
        return f

    def handle_event(self, event, *args):
        if event not in self._handlers:
            logging.debug(f"event type `{event}` not handled")
            return

        handlers = self._handlers.get(event)

        if not handlers:
            logging.debug(
                f"no handler registered for event type `{event}`"
            )
            return

        for handler in handlers:
            if args:
                handler(*args)
            else:
                handler()



from json import loads, dumps

from .._event_types import EventTypes


class ChatUser:
    def __init__(self, nick, features):
        self.nick = nick
        self.features = features

    @classmethod
    def from_ws_messsage(cls, msg):
        return cls(nick=msg.get('nick'), features=msg.get('features'))

    def __repr__(self):
        return f"ChatUser(nick='{self.nick}')"


class Message:
    def __init__(self, msg):
        split = msg.split()
        self.event = split[0]
        payload = ' '.join(split[1:])
        try:
            self.payload = loads(payload)
        except:
            self.payload = payload

    def __repr__(self):
        obj = self.__dict__.copy()
        if type(self).__name__ != Message.__name__:
            del obj['payload']
        return dumps(obj, default=lambda o: o.__dict__, ensure_ascii=False)

    @classmethod
    def parse(cls, msg):
        event = msg.split()[0]
        if event == EventTypes.SERVED_CONNECTIONS:
            return ServedConnections(msg)
        if event == EventTypes.USER_JOINED:
            return UserJoined(msg)
        if event == EventTypes.USER_QUIT:
            return UserQuit(msg)
        if event == EventTypes.BROADCAST:
            return Broadcast(msg)
        if event == EventTypes.CHAT_MESSAGE:
            return ChatMessage(msg)
        if event == EventTypes.WHISPER:
            return Whisper(msg)
        if EventTypes.is_moderation_event(event):
            return ModerationMessage(msg)
        return cls(msg)

    @property
    def json(self):
        obj = self.__dict__.copy()
        if type(self).__name__ != Message.__name__:
            del obj['payload']
        return dumps(obj, default=lambda o: o.__dict__, indent=4, ensure_ascii=False)


class ServedConnections(Message):
    def __init__(self, msg):
        super().__init__(msg)
        self.count = self.payload.get('connectioncount')
        self.users = [
            ChatUser.from_ws_messsage(u) for u in self.payload.get('users')
        ]


class UserJoined(Message):
    def __init__(self, msg):
        super().__init__(msg)
        self.user = ChatUser.from_ws_messsage(self.payload)
        self.timestamp = self.payload.get('timestamp')/1000.


class UserQuit(Message):
    def __init__(self, msg):
        super().__init__(msg)
        self.user = ChatUser.from_ws_messsage(self.payload)
        self.timestamp = self.payload.get('timestamp')/1000.


class Broadcast(Message):
    def __init__(self, msg):
        super().__init__(msg)
        self.timestamp = self.payload.get('timestamp')/1000.
        self.content = self.payload.get('data')


class ChatMessage(Message):
    def __init__(self, msg):
        super().__init__(msg)
        self.user = ChatUser.from_ws_messsage(self.payload)
        self.timestamp = self.payload.get('timestamp')/1000.
        self.content = self.payload.get('data')


class Whisper(Message):
    def __init__(self, msg):
        super().__init__(msg)
        self.user = ChatUser.from_ws_messsage(self.payload)
        self.message_id = self.payload.get('messageid')
        self.timestamp = self.payload.get('timestamp')/1000.
        self.content = self.payload.get('data')


class ModerationMessage(Message):
    def __init__(self, msg):
        super().__init__(msg)
        self.moderator = ChatUser.from_ws_messsage(self.payload)
        self.timestamp = self.payload.get('timestamp')/1000.
        user, *sentence = self.payload.get('data').split()
        self.affected_user = user
        self.sentence = ' '.join(sentence)


class SubOnly(Message):
    def __init__(self, msg):
        super().__init__(msg)
        self.moderator = ChatUser.from_ws_messsage(self.payload)
        self.timestamp = self.payload.get('timestamp')/1000.
        # on or off
        self.mode = self.payload.get('data')



from setuptools import setup, find_packages
from pathlib import Path


long_description = Path('README.md').read_text()

setup(
    name='dgg-chat',
    packages=find_packages(),
    version='0.1.2-alpha',
    license='MIT',
    description='A package that lets you do stuff in dgg chat',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Gabriel Jablonski',
    author_email='contact@gabrieljablonski.com',
    url='https://github.com/gabrieljablonski/dgg-chat',
    download_url='https://github.com/gabrieljablonski/dgg-chat/archive/v0.1.0-alpha.tar.gz',    # I explain this later on
    keywords=['chat-bot', 'chat', 'destinygg', 'dgg'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
    ],
    python_requires='>=3.8',
    install_requires=['websocket_client', 'requests'],
)
